#
# AitCS - Array in the Command's Shadow
#
# Stuart Cassoff
#
# Spring 2018
# Version 1.1
#
# Winter 2020
# Version 1.2
#
# Spring 2020
# Version 1.3
# Version 1.4
# Version 1.5
# Version 1.6
#

package require Tcl 8.6

namespace eval ::aitcs {

proc LICENSE {} { return [dict create type ISC text [string range {

Copyright (c) 2018,2020 Stuart Cassoff <stwo@users.sourceforge.net>

Permission to use, copy, modify, and distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

} 2 end-2]] }; #esnecil

proc ABOUT {} {
	return [dict create name aitcs title AitCS descr "Array in the Command's Shadow" author {Stuart Cassoff} when {Spring 2020} version [package provide aitcs]]
}

proc randstr {{length {6}}} {
	return [subst [string repeat {[format %c [expr {97 + int(rand() * 26)}]]} $length]]
}

proc randvar {name variable_namespace} {
	return [string trimright $variable_namespace :]::__${name}__[randstr]_
}

proc name {name {variable_namespace ""}} {
	set vns $variable_namespace
	if {$vns eq ""} { set vns [uplevel 1 {namespace current}] }
	return [randvar $name $vns]
}

proc new {{name ""} {command_namespace ""} {variable_namespace ""} args} {
	set rand [expr {$name eq ""}]
	if {$rand} { set name [randstr] }
	set cns $command_namespace
	set vns $variable_namespace
	if {$vns eq ""} { set vns [uplevel 1 {namespace current}] }
	if {$cns eq ""} { set cns $vns }
	set cns [string trimright $cns :]:
	if {[llength [info commands $cns:$name]] > 0} {
		return -code error "$cns:$name already exists"
	}
	set vns [string trimright $vns :]:
	set myns [string trimright [namespace current] :]:
	set varname [name $name $vns]
	interp alias {} $cns:$name    {} $myns:vset     $varname
	interp alias {} $cns:$name?   {} $myns:vexists  $varname
	interp alias {} $cns:$name-   {} $myns:vunset   $varname
	interp alias {} $cns:$name^   {} $myns:vgwd     $varname
	interp alias {} $cns:$name&   {} $myns:vref     $varname
	interp alias {} $cns:$name=   {} $myns:vsetto   $varname
	interp alias {} $cns:$name+   {} $myns:vlappend $varname
	interp alias {} $cns:$name+=  {} $myns:vincr    $varname
	interp alias {} $cns:$name.   {} $myns:vaitcs   $varname
	interp alias {} $cns:$name!   {} upvar #0 $varname $name
	interp alias {} $cns:$name&&  {} return -level 0 $cns:$name
	interp alias {} $cns:$name--- {} eval "
		interp alias {} $cns:$name    {}
		interp alias {} $cns:$name?   {}
		interp alias {} $cns:$name-   {}
		interp alias {} $cns:$name^   {}
		interp alias {} $cns:$name&   {}
		interp alias {} $cns:$name=   {}
		interp alias {} $cns:$name+   {}
		interp alias {} $cns:$name+=  {}
		interp alias {} $cns:$name.   {}
		interp alias {} $cns:$name!   {}
		interp alias {} $cns:$name&&  {}
		interp alias {} $cns:$name--- {}
		unset $varname
	"
	array set $varname $args
	return [expr {$rand ? "$cns:$name" : $varname}]
}

proc vunset {varname elem args} {
	foreach e [linsert $args 0 $elem] { array unset $varname $e }
	return ""
}

proc vref {varname {elem ""}} {
	if {$elem eq ""} { return $varname }
	return ${varname}($elem)
}

proc vexists {varname elem {myVar ""}} {
	upvar 1 $varname v
	if {![info exists v($elem)]} { return 0 }
	if {$myVar ne ""} {
		upvar 1 $myVar var
		set var $v($elem)
	}
	return 1
}

proc vset {varname elem args} {
	upvar 1 $varname v
	if {[llength $args] > 0} { array set v [linsert $args 0 $elem] }
	return $v($elem)
}

proc vsetto {varname elem elemOrVal args} {
	upvar 1 $varname v
	set a [linsert $args 0 $elem $elemOrVal]
	set val [lindex $a end]
	foreach e [lrange $a 0 end-1] { set v($e) $val }
	return $val
}

proc vlappend {varname elem args} {
	upvar 1 $varname v
	lappend v($elem) {*}$args
	return $v($elem)
}

proc vincr {varname elem {incr {1}}} {  
        upvar 1 $varname v
        incr v($elem) $incr
	return $v($elem)
}

proc vgwd {varname elem {default ""} {myVar ""}} {
	if {$myVar ne ""} { upvar 1 $myVar var }
	set var [expr {[vexists $varname $elem] ? [vset $varname $elem] : $default}]
}

proc vaitcs {varname elem args} {
	if {[set op [string range $elem end-1 end]] in [list += &&]} {
		set elem [string range $elem 0 end-2]
	} elseif {[set op [string index $elem end]] in [list ? - ^ & = + . ! :]} {
		set elem [string range $elem 0 end-1]
	} elseif {[set op [string range $elem end-2 end]] eq "---"} {
		set elem [string range $elem 0 end-3]
	} else {
		set op ""
	}
	set aitcs [vset $varname $elem]
	uplevel 1 $aitcs$op $args
}

namespace ensemble create -prefixes no -command aitcs -subcommands {name new}
namespace export aitcs

package provide aitcs 1.6

}; # namespace eval ::aitcs

# EOF
